//*** The code for the simulations of the article 
//*** "Combined influence of nutrient supply level and tissue mechanical properties on benign tumor growth as revealed by mathematical modeling"
//*** by Maxim Kuznetsov
//*** journal "Mathematics", Special Issue "Mathematical Modelling in Biomedicine II"
//*** 2021
//*** This code works fine in Dev-C++, version 5.5.2 
//*** Note, that some small parts of the code may need to be revised depending on the utilized software 

//***********************************************************************************************************************************************************
//********************************* INTRODUCTION ************************************************************************************************************
//***********************************************************************************************************************************************************

//********* Initializing needful things

#include <stdio.h>
#include <tchar.h>
#include <windows.h>
#include <ctime>
#include <iostream>
#include <cmath>
#include <conio.h>
#include <fstream>

using namespace std;

//********* MODEL PARAMETERS

int HCinf = 0; // set 1 for infinite hydraulic conductivity limit

// For the full model and the infinite hydraulic conductivity limit:

double B = 0.03; // tumor cells proliferation rate
double M = 0.003; // tumor cells death rate
double P = 4.; // nutrient supply level
double Dg = 100.; // glucose diffusion coefficient
double Qp = 24.; // rate of glucose consumption by proliferating tumor cells
double Qq = 0.5;  // rate of glucose consumption by quiescent tumor cells
double c0 = 0.8; // initial fraction of cells
double gp = 0.1; // critical level of glucose for tumor cell proliferation
double gd = 0.01; // critical level of glucose for tumor cell survival
double eps = 1000.; // Heaviside smoothing parameter
double R0T = 100.; // initial radius of tissue region

// Only for the full model (HCinf !=1)

double scr = 15.; // critical stress for cell proliferation
double Kmu = 3.; // hydraulic conductivity
double k = 100.; // cells stress function coefficient
double cs = 0.3; // minimum cell-sensing cell fraction

# define PI 3.14159265358979323846  // pi

//********* SIMULATION PARAMETERS

#define size 5000. // the radius of a simulation region, in 10^(-2) cm

double tau = 0.00001; // initial time step in full model -- will be adjusted 
double tauHCinf = 0.001; // will be set for infinite hydraulic conductivity limit
double tauAdjust;
double I0NonMonIndexMin=1.e-5; // lower value of this parameter...
double tauAdjustParameter=100; // and greater value of this parameter provide more accurte calculations
								// under these values artificial cells may be kept in tumor center in case of fluid-filled core
								// however, they do not noticeably affect tumor growth curve, and the simulation time is kept tolerable
#define h 0.05 // space step
double tauMax = 200.*h*h / (2.*Dg);

#define N (int(size/h)) // number of points in space, minus one
#define tau_adjust_frequency 0.1 // how often time step will be adjusted
#define rsn_frequency 0.1 // how often the tumor radius, speed of tumor front and number of tumor cells will be recorded
#define output_frequency 24. // how often the distributions of varibles will be recorded
FILE *results;
#define inception_width 1. // the radius of inital tumor cell population
#define inception_width_int (int)(inception_width/h) // the size, in cell grids, of inital tumor cell population

//********* COUNTERS

double t = 0.; // main
double ta = 0.; // for adjustment of time step
double ti = 0.; // for output of the distributions of varibles
double tr = 0.; // for output of the tumor radius, speed of tumor front and number of tumor cells
int d = 0; // for serial numbers of the files with distributions of varibles
int j; // for estimation of Rd and Rp

//********* MODEL VARIABLES

double n0[N + 1], nr[N + 1], n1[N + 1]; // tumor cells
double h0[N + 1], hr[N + 1], h1[N + 1]; // normal cells
double g0[N + 1], gr[N + 1], g1[N + 1]; // glucose
double G0[N + 1], GR[N + 1], G1[N + 1]; // glucose -- for easy solution in spherically-symmetrical case
double nh0[N + 1], nh1[N + 1]; // all cells -- for convenient solution of convection equation
double sigma[N + 1]; // stress

//********* PROCEDURES AND FUNCTIONS

void Begin(); // complex of actions for the beginning, ...
void Cycle_Ending(); // ...ending of main cycle, ...
void End(); // ...and ending of the program execution

void TauAdjust(); // for adjustment of time step

void Convection(); // solving the equation for cells convection
void Stress(double* nh); // computing stress
void ConvSpeed(double* nh, double* II); // computing convective speed
void ConvSpeedHCinf(double* nn, double* sp, double* sd, double* II); // computing convective speed in case of infinite hydraulic conductivity limit

void Diffusion_glucose(); //  solving the equation for diffusion of glucose

void TMAc(double apr, double bpr, double cpr, double* F, double* prog,
	double muleft, double hileft, double muright, double hiright); // tridiagonal matrix algorithm

void Euler(); // Euler method
double Growthn(double nn, double sp, double sd, double ss, double II, double rr); // kinetic terms for tumor cells
double Growthh(double hh, double II, double rr); // kinetic terms for normal cells
double Growthg(double nn, double hh, double gg, double sp, double sd, double ss); // kinetic terms for glucose in tumor
double Growthgh(double hh, double gg); // kinetic terms for glucose in normal tissue (eases computation)

double thetap(double gg); // function that defines the rate of proliferation of tumor cells depending on the level of glucose
double thetad(double gg); // function that defines the rate of death of tumor cells depending on the level of glucose
double thetas(double ss); // function that defines the rate of proliferation of tumor cells depending on the solid pressure
#define Nm 10000 // number of precalculated values of theta functions
double thetapPre[Nm + 1]; // values for theta_p(g) are precalculated for optimization purposes
double thetadPre[Nm + 1]; // values for theta_d(g) are precalculated for optimization purposes
double thetasPre[Nm + 1]; // values for theta_s(P) are precalculated for optimization purposes
double thetapg[N + 1];
double thetadg[N + 1];
double thetass[N + 1];

void Profile_output(); // output of distributions of variables
void Radius_Speed_CellNumber_output(); // output of the tumor radius, whole tissue radius, speed of tumor front and number of tumor cells
void Radius_estimation(); // estimation of tumor radius and whole tissue radius
void Speed_estimation(); // estimation of tumor front speed

//********* AUXILIARY VARIABLES

////// for glucose diffusion

double kg = 2 * h*h / (tau*Dg);
double aprg = 1.;
double bprg = 2. + kg;
double cprg = 1.;
double muleftg, hileftg, murightg, hirightg;
double Fg[N + 1];
double progg[N + 1];
double pp[N + 1], qq[N + 1];

////// for cells convection

double Sgn(double x);
double I0[N + 2], I0_not_corr[N + 3];
double I0Max;
double I0NonMonIndex;
double Q_plus[N + 2], Q_minus[N + 2];
double delta_nh[N + 2];
double one_eighth[N + 2];
double f_c_nh[N + 2];

int NN = inception_width_int; // number of point in space where tumor cells are
int NH = (int(R0T / h)); // number of point in space where tumor and normal cells are
double hNN = 0.; // the variable size of the last grid element, NN, where tumor cells are
double hNH = 0.; // the variable size of the last grid element, NH, where normal cells are
double hNNnew;
double VNNnew;
double hNHnew;
double VNHnew;

////// for tumor radius, speed of tumor front and number of tumor cells

double RadTis; // current whole tissue radius, in mm
double Rad; // current tumor radius, in mm
double Rad_prev; // previously measured tumor radius, in mm
double Rp; // distance from tumor center to the border of quiescent and proliferating zones
double Rd; // distance from tumor center to the border of dying cells region and quiescent zone
double T_prev; // time of previous measurement of tumor radius
double V; // tumor front speed
double NTC, NTCr, NTCcm1; // relative number of tumor cells
double HTCr, HTCcm1; // relative number of normal cells -- for solution of convective equation
double T_est_fin = 0.; // auxiliary variable for program stop
double V_prev; // auxiliary variable for program stop

////// for work with files for distribution of variables

char File_name1[80], num[10], radius[8];

//***********************************************************************************************************************************************************
//********************************* MAIN CYCLE **************************************************************************************************************
//***********************************************************************************************************************************************************

int main()
{
	Begin();
	CreateDirectoryA("Profiles", NULL);

	// the program runs until tumor grows no more, that is checked in Speed_estimation()
	while (1 == 1)
	{
		t = t + tau;
		Radius_estimation();
		Speed_estimation();

		Euler();
		Convection();
		Diffusion_glucose();

		if (t >= ti - 0.00001) // a trifle is subtracted to overcome the inevitable minor machine errors during arithmetic operations
			Profile_output();
		if (t >= tr)
			Radius_Speed_CellNumber_output();

		if (HCinf != 1 && t >= ta - 0.00001)
			TauAdjust();

		Cycle_Ending();
	}
}

//***********************************************************************************************************************************************************
//********************************* BEGINNING OF SIMULATION *************************************************************************************************
//***********************************************************************************************************************************************************

void Begin()
{
	//********* PRECALCULATING thetas
	for (int i = 0; i <= Nm; i++)
	{
		thetapPre[i] = thetap(1.*i / Nm);
		thetadPre[i] = thetad(1.*i / Nm);
	}

	if (HCinf == 1)
		for (int i = 0; i <= NH; i++)
			thetass[i] = 1.;
	else
		for (int i = 0; i <= Nm; i++)
			thetasPre[i] = thetas(-k*(1-c0)+2.*k*(1-c0)*i / Nm);
	
	//********* CREATING FILE FOR TUMOR RADIUS, FRONT SPEED AND CELL NUMBER
	if ((results = fopen("results.txt", "wt")) == NULL)
		fprintf(stderr, "Cannot open file results.txt\n\n");

	fprintf(results, "Time(h)\tRad(mm)\tSpeed\tCells\t\tRadTis(mm)\tRq\tRd\ttau\t\tI0NonMonIndex\n\n");
	fflush(results);

	//********* INITIAL CONDITIONS

	for (int i = 0; i <= NH; i++)
	{
		n0[i] = 0.;
		h0[i] = c0;
		g0[i] = 1.;
		nh0[i] = c0;
	}

	for (int i = 0; i <= NN - 1; i++)
	{
		n0[i] = c0;
		h0[i] = 0.;
	}

	n0[NN] = c0;
	h0[NN] = c0;

	for (int i = 0; i <= N; i++) // this is made for easy output of the initial conditions
	{
		n1[i] = n0[i];
		h1[i] = h0[i];
		g1[i] = g0[i];
	}

	Radius_estimation();

	//********* COUNTING TUMOR CELLS
	NTC = 0.;

	for (int i = 0; i <= NN - 1; i++)
		NTC = NTC + ((1.*i + 0.5)*h)*((1.*i + 0.5)*h)*h*n1[i];
	//hNN is yet zero

	NTC = 4.*PI*300.*NTC;

	Profile_output();
}

//***********************************************************************************************************************************************************
//********************************* OUTPUT OF DISTRIBUTIONS OF VARIABLES ************************************************************************************
//***********************************************************************************************************************************************************

void Profile_output()
{
	//********* CREATING A FILENAME WITH SERIAL NUMBER

	d = (int)((ti + 0.00001) / output_frequency);
	_itoa(d, num, 10);

	if (d < 10)
		strcpy(File_name1, "res-d0000"); // designed for maximum a thousand of files
	else if (d < 100)
		strcpy(File_name1, "res-d000");
	else if (d  < 1000)
		strcpy(File_name1, "res-d00");
	else if (d  < 10000)
		strcpy(File_name1, "res-d0");
	else
		strcpy(File_name1, "res-d");

	strcat(File_name1, num);

	if (RadTis<1.)
	{
		_gcvt(RadTis, 1, radius);
		if (radius[1] == '\0')
			radius[1] = '0';
	}
	else if (RadTis<10.)
	{
		_gcvt(RadTis, 2, radius);
		if (radius[2] == '\0')
			radius[2] = '0';
	}
	else
	{
		_gcvt(RadTis, 3, radius);// in mm, designed for no more than 1000 mm
		if (radius[3] == '\0')
			radius[3] = '0';
	}

	strcat(File_name1, "_RadTis(mm)=");

	strcat(File_name1, radius);

	strcat(File_name1, ".txt");

	//********* PRINTING THE SERIAL NUMBER, RADIUS, SPEED OF TUMOR FRONT AND NUMBER OF TUMOR CELLS ON A SCREEN

	printf("day %i, radius = %.3f mm, speed = %.3f mm/week, tissue radius = %.3f mm, cell number = %.3e\n", d, Rad, V, RadTis, NTC);

	//********* OUTPUT OF DISTRIBUTIONS OF VARIABLES INTO A FILE

	FILE * out;
	if ((out = fopen(File_name1, "wt")) == NULL)
		fprintf(stderr, "Cannot open file\n");


	fprintf(out, "Len(mm)\t\tn\t\th\t\tg\t\tI\t\tsigma\n\n");

	for (int i = 0; i <= NN - 1; i++)
		fprintf(out, "%.6f\t%.7e\t%.7e\t%.7e\t%.7e\t%.5f\n", (1.*i + 0.5)*h / 10., n1[i], h1[i], g1[i], I0[i], sigma[i]);

	fprintf(out, "%.6f\t%.7e\t%.7e\t%.7e\t%.7e\t%.5f\n", ((1.*NN)*h + hNN) / 10., n1[NN], h1[NN], g1[NN - 1] + (g1[NN + 1] - g1[NN - 1])*((h / 2.) + hNN) / (2.*h), I0[NN], sigma[NN]);

	for (int i = NN + 1; i <= NH - 1; i++)
		fprintf(out, "%.6f\t%.7e\t%.7e\t%.7e\t%.7e\t%.5f\n", (1.*i + 0.5)*h / 10., n1[i], h1[i], g1[i], I0[i], sigma[i]);

	fprintf(out, "%.6f\t%.7e\t%.7e\t%.7e\t%.7e\t%.5f\n", ((1.*NH)*h + hNH) / 10., n1[NH], h1[NH], g1[NH], I0[NH], sigma[NH]);

	fclose(out);

	//********* SWITCHING THE COUNTER FOR OUTPUT OF DISTRIBUTION OF VARIABLES 

	ti = ti + output_frequency;
}

//***********************************************************************************************************************************************************
//********************************* ESTIMATION OF RADIUS AND SPEED ******************************************************************************************
//***********************************************************************************************************************************************************

//********* ESTIMATION OF RADIUS

void Radius_estimation()
{
	Rad = (NN*h + hNN) / 10.;
	RadTis = (NH*h + hNH) / 10.;

	if (g0[0] > gp)
		Rp = 0.;
	else
	{
		j = 0;
		while (g0[j] < gp)
			j++;
		Rp = j*h;
	}

	if (g0[0] > gd)
		Rd = 0.;
	else
	{
		j = 0;
		while (g0[j] < gd)
			j++;
		Rd = j*h;
	}
}

//********* ESTIMATION OF TUMOR FRONT SPEED

void Speed_estimation()
{
	if (t - T_prev > 5. || abs(Rad - Rad_prev) > 0.2)
	{
		V = 7.*24.*(Rad - Rad_prev) / (t - T_prev); // in mm/week
		T_prev = t;
		Rad_prev = Rad;
	}

	if (t - T_est_fin > 10.*24.)
	{
		if (abs(V) < 1.e-5 && abs(V_prev) < 1.e-5) // growth stop
		{
			cout << "\n\nEnding.";
			_getch();
			End();
		}
		else
			T_est_fin = t;

		V_prev = V;
	}
}

//***********************************************************************************************************************************************************
//********************************* OUTPUT OF THE TUMOR RADIUS, SPEED OF TUMOR FRONT AND NUMBER OF TUMOR CELLS **********************************************
//***********************************************************************************************************************************************************

void Radius_Speed_CellNumber_output()
{
	tr = tr + rsn_frequency;

	//********* COUNTING TUMOR CELLS

	NTC = 0.;

	for (int i = 0; i <= NN - 1; i++)
		NTC = NTC + ((1.*i + 0.5)*h)*((1.*i + 0.5)*h)*h*n1[i];

	NTC = NTC + (NN*h + hNN/2.)*(NN*h + hNN/2.)*hNN*n1[NN];

	NTC = 4.*PI*300.*NTC;

	//********* OUTPUT

	fprintf(results, "%.1f\t%.5f\t%.5f\t%.3e\t%.5f\t%.3f\t%.3f\t%.3e\t%.3e\n", t, Rad, V, NTC, RadTis, 1.*Rp/10., 1.*Rd/10., tau, I0NonMonIndex);
	fflush(results);
}

//***********************************************************************************************************************************************************
//********************************* SOLUTION OF KINETIC TERMS ***********************************************************************************************
//***********************************************************************************************************************************************************

//********* FUNCTIONS THAT DEFINE THE BEHAVIOR OF TUMOR CELLS DEPENDING ON THE LEVEL OF GLUCOSE AND SOLID STRESS

double thetap(double gg)
{
	return (0.5*(1. + tanh(eps*(gg-gp))));
}

double thetad(double gg)
{
	return (0.5*(1. + tanh(eps*(gd-gg))));
}

double thetas(double ss)
{
	return (0.5*(1. + tanh(eps*(scr-ss))));
}

//********* KINETIC FUNCTIONS

double Growthn(double nn, double sp, double sd, double ss, double II, double rr) // for tumor cells
{
	return ((B*sp*ss - M*sd - 2.*II / rr)*nn);
}

double Growthh(double hh, double II, double rr) // for normal cells
{
	return (-2.*II*hh / rr);
}

double Growthg(double nn, double hh, double gg, double sp, double sd, double ss) // for glucose in tumor
{
	return (P*hh*(1. - gg) - (Qp*sp*ss + Qq*((1. - sp)*ss + (1. - ss))*(1 - sd))*nn);
}

double Growthgh(double hh, double gg) // for glucose in normal tissue
{
	return (P*hh*(1. - gg));
}

//********* EULER PROCEDURE

void Euler()
{
	for (int i = 0; i <= NN - 1; i++)
	{
		thetapg[i] = thetapPre[(int)(g0[i] * Nm + 0.5)];
		thetadg[i] = thetadPre[(int)(g0[i] * Nm + 0.5)];
	}

	thetapg[NN] = thetapPre[(int)(g0[NN] * Nm + 0.5)];
	thetadg[NN] = thetadPre[(int)(g0[NN] * Nm + 0.5)];

	// computing convective speed
	if (HCinf == 1)
		ConvSpeedHCinf(n0, thetapg, thetadg, I0);
	else
	{
		// redefinition of variables for convenience of computing convective speed
		for (int i = 0; i <= NN - 1; i++)
			nh0[i] = n0[i];

		nh0[NN] = (n0[NN] * ((NN*h + hNN)*(NN*h + hNN)*(NN*h + hNN) - (NN*h)*(NN*h)*(NN*h)) + h0[NN] * (((NN + 1)*h)*((NN + 1)*h) *((NN + 1)*h) - (NN*h + hNN)*(NN*h + hNN)*(NN*h + hNN))) / (((NN + 1)*h)*((NN + 1)*h) *((NN + 1)*h) - (NN*h)*(NN*h)*(NN*h));

		for (int i = NN + 1; i <= NH; i++)
			nh0[i] = h0[i];

		Stress(nh0);
		ConvSpeed(nh0, I0);

		for (int i = 0; i <= NN - 1; i++)
			thetass[i] = thetasPre[(int)((sigma[i] + k*(1-c0)) * Nm / (2.*k*(1-c0))  + 0.5)];
		thetass[NN] = thetasPre[(int)((sigma[NN] + k*(1-c0)) * Nm / (2.*k*(1-c0))  + 0.5)];
	}

#pragma loop(hint_parallel(6))  
#pragma loop(ivdep) 
	for (int i = 0; i <= NN - 1; i++)
	{
		nr[i] = n0[i] + tau*Growthn(n0[i], thetapg[i], thetadg[i], thetass[i], I0[i], (1.*i + 0.5)*h);
		gr[i] = g0[i] + tau*Growthg(n0[i], h0[i], g0[i], thetapg[i], thetadg[i], thetass[i]);
	}

	nr[NN] = n0[NN] + tau*Growthn(n0[NN], thetapg[NN], thetadg[NN], thetass[NN], I0[NN], (1.*NN + 0.5)*h);
	hr[NN] = h0[NN] + tau*Growthh(h0[NN], I0[NN], (1.*NN + 0.5)*h);
	gr[NN] = g0[NN] + tau*Growthg(n0[NN] * ((NN*h + hNN)*(NN*h + hNN)*(NN*h + hNN) - (NN*h)*(NN*h)*(NN*h)) / (((NN + 1)*h)*((NN + 1)*h) *((NN + 1)*h) - (NN*h)*(NN*h)*(NN*h)), h0[NN] * (((NN + 1)*h)*((NN + 1)*h) *((NN + 1)*h) - (NN*h + hNN)*(NN*h + hNN)*(NN*h + hNN)) / (((NN + 1)*h)*((NN + 1)*h) *((NN + 1)*h) - (NN*h)*(NN*h)*(NN*h)), g0[NN], thetapg[NN], thetadg[NN], thetass[NN]);

	for (int i = NN + 1; i <= NH - 1; i++)
	{
		hr[i] = h0[i] + tau*Growthh(h0[i], I0[i], (1.*i + 0.5)*h);
		gr[i] = g0[i] + tau*Growthgh(h0[i], g0[i]);
	}

	hr[NH] = h0[NH] + tau*Growthh(h0[NH], I0[NH], NH*h + hNH / 2.);
	gr[NH] = g0[NH] + tau*Growthgh(h0[NH], g0[NH]);
}

//***********************************************************************************************************************************************************
//********************************* CONVECTION OF CELLS *****************************************************************************************************
//***********************************************************************************************************************************************************

//********* PROCEDURE FOR COMPUTING CONVECTIVE SPEED

void Stress(double* nh)
{
	for (int i = 0; i <= NH; i++)
	{
		if (nh[i] >= c0)
			sigma[i] = k*(nh[i]-c0);

		else if (nh[i] > cs)
			sigma[i] = k*(nh[i]-c0)*(nh[i]-cs)*(nh[i]-cs)/((c0-cs)*(c0-cs));
		else
			sigma[i] = 0.;
	}
}

void ConvSpeed(double* nh, double* II)
{
	II[0] = Kmu*(sigma[0] - sigma[1]) / (2.*h);

#pragma loop(hint_parallel(6))  
#pragma loop(ivdep) 
	for (int i = 1; i <= NH - 2; i++)
		II[i] = Kmu*(sigma[i - 1] - sigma[i + 1]) / (2.*h);

	II[NH - 1] = II[NH - 3] + 2.*(II[NH - 2] - II[NH - 3]);

	II[NH] = II[NH - 2] + (II[NH - 1] - II[NH - 2])*(1.5*h + hNH) / h;
}

void ConvSpeedHCinf(double* nn, double* sp, double* sd, double* II)
{
	II[0] = (B * sp[0] - M * sd[0])*nn[0] * 0.25*h*0.25*h*0.5*h;

	for (int i = 1; i <= NN - 1; i++)
		II[i] = II[i - 1] + (B*sp[i - 1] - M*sd[i - 1])*nn[i - 1] * ((1.*i - 0.25)*h)*((1.*i - 0.25)*h)*0.5*h + (B*sp[i] - M*sd[i])*nn[i] * ((1.*i + 0.25)*h)*((1.*i + 0.25)*h)*0.5*h;

	II[NN] = II[NN - 1] + (B*sp[NN - 1] - M*sd[NN - 1])*nn[NN - 1] * ((1.*NN - 0.25)*h)*((1.*NN - 0.25)*h)*0.5*h + (B*sp[NN] - M*sd[NN])*nn[NN] * (NN*h+0.5*hNN)*(NN*h + 0.5*hNN)*hNN;
	
	for (int i = NN + 1; i <= NH; i++)
		II[i] = II[i - 1]; // nn is further 0.

	for (int i = 0; i <= NH - 1; i++)
		II[i] = II[i] / (((1.*i + 0.5)*h)*((1.*i + 0.5)*h)) / c0;

	II[NH] = II[NH] / ((NH*h + hNH)*(NH*h + hNH)) / c0;
}

//********* SIGN FUNCTION

double Sgn(double x)
{
	if (x>0.)
		return (1.);
	else if (x == 0.)
		return (0.);
	else
		return (-1.);
}

//********* PROCEDURE FOR CELLS' CONVECTION (more precisely, its "linear" part)

void Convection()
{
	for (int i = 0; i <= NN - 1; i++)
		nh0[i] = nr[i];

	nh0[NN] = nr[NN] * (hNN / h) + hr[NN] * ((h - hNN) / h);

	for (int i = NN + 1; i <= NH; i++)
		nh0[i] = hr[i];


	if (HCinf == 1)
		ConvSpeedHCinf(n0, thetapg, thetadg, I0_not_corr);
	else
	{
		Stress(nh0);
		ConvSpeed(nh0, I0_not_corr);
	}

	// adjusting speed
	if (I0_not_corr[0] >= 0.)
		I0[0] = (1. - (I0_not_corr[0] * tau / (2.*h)))*I0_not_corr[0] + (I0_not_corr[0] * tau / (2.*h))*I0_not_corr[1];
	else
		I0[0] = (1. + (I0_not_corr[0] * tau / (2.*h)))*I0_not_corr[0] + (I0_not_corr[0] * tau / (2.*h))*I0_not_corr[0];

	for (int i = 1; i <= NH - 1; i++)
	{
		if (I0_not_corr[i] >= 0.)
			I0[i] = (1. - (I0_not_corr[i] * tau / (2.*h)))*I0_not_corr[i] + (I0_not_corr[i] * tau / (2.*h))*I0_not_corr[i + 1];
		else
			I0[i] = (1. + (I0_not_corr[i] * tau / (2.*h)))*I0_not_corr[i] - (I0_not_corr[i] * tau / (2.*h))*I0_not_corr[i - 1];
	}

	if (I0_not_corr[NH] < 0.)
		I0[NH] = (1. + (I0_not_corr[NH] * tau / (h + hNH)))*I0_not_corr[NH] - (I0_not_corr[NH] * tau / (h + hNH))*I0_not_corr[NH - 1];
	else
		I0[NH] = I0_not_corr[NH];

	// defining auxiliary variables
	for (int i = 0; i <= NH - 2; i++)
		Q_plus[i] = (0.5 - I0[i] * tau / h) / (1. + (I0[i + 1] - I0[i])*tau / h);
	for (int i = 1; i <= NH - 1; i++)
		Q_minus[i] = (0.5 + I0[i] * tau / h) / (1. - (I0[i - 1] - I0[i])*tau / h);

	Q_plus[NH - 1] = (0.5*h - I0[NH - 1] * tau) / (0.5*(h + hNH) + (I0[NH] - I0[NH - 1])*tau);

	Q_minus[NH] = (0.5*hNH + I0[NH] * tau) / (0.5*(h + hNH) + (I0[NH] - I0[NH - 1])*tau);

	// calculating the result of the transport stage
	nh1[0] = 0.5*Q_plus[0] * Q_plus[0] * (nh0[1] - nh0[0]) + Q_plus[0] * nh0[0] + 0.5*nh0[0];

#pragma loop(hint_parallel(6))  
#pragma loop(ivdep) 
	for (int i = 1; i <= NH - 2; i++)
		nh1[i] = 0.5*Q_minus[i] * Q_minus[i] * (nh0[i - 1] - nh0[i]) + 0.5*Q_plus[i] * Q_plus[i] * (nh0[i + 1] - nh0[i]) + (Q_plus[i] + Q_minus[i])*nh0[i];

	nh1[NH - 1] = 0.5*Q_minus[NH - 1] * Q_minus[NH - 1] * (nh0[NH - 2] - nh0[NH - 1]) + 0.5*Q_plus[NH - 1] * Q_plus[NH - 1] * (nh0[NH] - nh0[NH - 1])*(h + hNH) / (2.*h) + (Q_plus[NH - 1] * (h + hNH) / (2.*h) + Q_minus[NH - 1])*nh0[NH - 1];

	nh1[NH] = c0;

	//********* ANTIDIFFUSION STAGE

	for (int i = 0; i <= NH - 1; i++)
		delta_nh[i] = nh1[i + 1] - nh1[i];

	delta_nh[NH] = 0. - nh1[NH];
	 
	for (int i = 0; i <= NH + 1; i++)
		one_eighth[i] = 0.125 + 0.5*(I0[i] * tau / h)*(I0[i] * tau / h);

	f_c_nh[0] = Sgn(delta_nh[0])*max(0., min((-1.)*delta_nh[0] * Sgn(delta_nh[0]), min(one_eighth[0] * abs(delta_nh[0]), delta_nh[1] * Sgn(delta_nh[0]))));

#pragma loop(hint_parallel(6))  
#pragma loop(ivdep) 
	for (int i = 1; i <= NH - 1; i++)
		f_c_nh[i] = Sgn(delta_nh[i])*max(0., min(delta_nh[i - 1] * Sgn(delta_nh[i]), min(one_eighth[i] * abs(delta_nh[i]), delta_nh[i + 1] * Sgn(delta_nh[i]))));

	nh1[0] = nh1[0] - f_c_nh[0] - Sgn(delta_nh[0])*max(0., min(delta_nh[1] * Sgn(delta_nh[0]), min((-1.)*one_eighth[1] * abs(delta_nh[0]), (-1.)*delta_nh[0] * Sgn(delta_nh[0]))));

	for (int i = 1; i <= NH - 1; i++)
		nh1[i] = nh1[i] - f_c_nh[i] + f_c_nh[i - 1];

	// REDISTRIBUTING n AND h
	
	for (int i = 0; i <= NN - 1; i++)
	{
		n1[i] = nh1[i];
		h1[i] = 0.;
	}
	
	//********* COUNTING "NORMAL CELLS" BEFORE CONVECTIVE STAGE (since this part of equation is solved in linear regime and this variable should be conserved)
	// it is made now since further hNN will probably change and NN may change

	if (R0T != size)
	{
		HTCr = hr[NN] * (h - hNN);

		for (int i = NN + 1; i <= NH - 1; i++)
			HTCr = HTCr + hr[i] * h;

		HTCr = HTCr + hr[NH] * hNH;
	}

	//********* COUNTING "TUMOR CELLS" BEFORE CONVECTIVE STAGE (since this part of equation is solved in linear regime and this variable should be conserved)

	NTCr = 0.;
	for (int i = 0; i <= NN - 1; i++)
		NTCr = NTCr + nr[i] * h;

	NTCr = NTCr + nr[NN] * hNN;

	//********* COUNTING "TUMOR CELLS" AFTER CONVECTIVE STAGE UP TO NN-1
	NTCcm1 = 0.;
	for (int i = 0; i <= NN - 1; i++)
		NTCcm1 = NTCcm1 + nh1[i] * h;

	hNNnew = (NTCr - NTCcm1) / nh1[NN];

	if (hNNnew < -1.*h)
	{
		cout << "\nError! hNNnew <-h";
		system("Pause");
		exit(0);
	}
	else if (hNNnew < 0.)
	{
		n1[NN] = 0.;
		h1[NN] = nh1[NN];
		NN--;
		n1[NN] = nh1[NN];
		h1[NN] = nh1[NN];
		hNN = h + hNNnew;
	}
	else if (hNNnew < h)
	{
		n1[NN] = nh1[NN];
		h1[NN] = nh1[NN];
		hNN = hNNnew;
	}
	else if (hNNnew < 2.*h)
	{
		n1[NN] = nh1[NN];
		h1[NN] = 0.;
		NN++;
		n1[NN] = nh1[NN];
		h1[NN] = nh1[NN];
		hNN = (NTCr - NTCcm1 - nh1[NN - 1] * h) / nh1[NN];
	}
	else
	{
		cout << "\nError! hNNnew >= 2.*h";
		system("Pause");
		exit(0);
	}

	if (R0T == size)
		nh1[NH] = c0;
	else
	{
		//********* COUNTING "NORMAL CELLS" AFTER CONVECTIVE STAGE UP TO NN-1
		
		HTCcm1 = nh1[NN] * (h - hNN);

		for (int i = NN + 1; i <= NH - 1; i++)
			HTCcm1 = HTCcm1 + nh1[i] * h;

		hNHnew = (HTCr - HTCcm1) / nh1[NH];

		if (hNHnew < -1.*h)
		{
			cout << "\nError! hNHnew <-h";
			system("Pause");
			exit(0);
		}
		else if (hNHnew < 0.)
		{
			nh1[NH] = 0.;
			NH--;
			nh1[NH] = c0;
			hNH = h + hNHnew;
		}
		else if (hNHnew < h)
		{
			nh1[NH] = c0;
			hNH = hNHnew;
		}
		else if (hNHnew < 2.*h)
		{
			nh1[NH] = (((0.5*Q_minus[NH] * Q_minus[NH] * (nh0[NH - 1] - nh0[NH]) + Q_minus[NH] * nh0[NH])*(0.5*(h + hNH)) + nh0[NH] * (h - 0.5*hNH - I0[NH] * tau))) / h;
			NH++;
			nh1[NH] = c0;
			g0[NH] = g0[NH - 1];
			hNH = hNHnew - h;
		}
		else
		{
			cout << "\nError! hNHnew >= 2.*h";
			system("Pause");
			exit(0);
		}

	}

	for (int i = NN + 1; i <= NH; i++)
	{
		n1[i] = 0.;
		h1[i] = nh1[i];
	}

	if (HCinf == 1) // to overcome numerical errors, especially near center
	{
		for (int i = 0; i <= NN; i++)
			n1[i] = c0;
		for (int i = NN; i <= NH; i++)
			h1[i] = c0;
	}
}

//***********************************************************************************************************************************************************
//********************************* DIFFUSION OF GLUCOSE ****************************************************************************************************
//***********************************************************************************************************************************************************

void Diffusion_glucose()
{
	for (int i = 0; i <= NH; i++)
	{
		G0[i] = g0[i] * (1.*i + 0.5)*h; // for easy solution in spherical symmetrical case
		GR[i] = gr[i] * (1.*i + 0.5)*h;
	}

	for (int i = 1; i <= NH - 1; i++)
		Fg[i] = (G0[i + 1] - 2.*G0[i] + G0[i - 1]) + kg * GR[i];

	muleftg = 0.;
	hileftg = 0.5 / 1.5;

	murightg = 0.;
	hirightg = (1.*NH + 0.5) / (1.*NH - 0.5);

	TMAc(aprg, bprg, cprg, Fg, progg, muleftg, hileftg, murightg, hirightg);

	for (int i = 0; i <= NH; i++)
		G1[i] = progg[i];

	for (int i = 0; i <= NH; i++)
		g1[i] = G1[i] / ((1.*i + 0.5)*h);
}

//********* TRIDIAGONAL MATRIX ALGORITHM

void TMAc(double apr, double bpr, double cpr, double* F, double* prog,
	double muleft, double hileft, double muright, double hiright)
{
	pp[1] = hileft;
	qq[1] = muleft;

	for (int i = 1; i <= NH - 1; i++)
	{
		pp[i + 1] = cpr / (bpr - pp[i] * apr);
		qq[i + 1] = (apr*qq[i] + F[i]) / (bpr - pp[i] * apr);
	}

	prog[NH] = (muright + hiright*qq[NH]) / (1. - pp[NH] * hiright);

	for (int i = NH - 1; i >= 0; i--)
		prog[i] = pp[i + 1] * prog[i + 1] + qq[i + 1];
}

//***********************************************************************************************************************************************************
//********************************* TIME STEP ADJUSTMENT ****************************************************************************************************
//***********************************************************************************************************************************************************

void TauAdjust()
{
	ta = ta + tau_adjust_frequency;

	// adjusting tau
	I0Max = 0.;
	I0NonMonIndex = 0;

	for (int i = NN + 1; i <= NH; i++)
		I0Max = max(I0Max, I0[i]);

	for (int i = NN+1; i <= NH-1; i++)
		if ((I0[i] - I0[i - 1]) * (I0[i + 1] - I0[i]) < -1.e-10)
			I0NonMonIndex= I0NonMonIndex - (I0[i] - I0[i - 1]) * (I0[i + 1] - I0[i]);

	tauAdjust = h / (2.*I0Max) / tauAdjustParameter;

	if (I0NonMonIndex > I0NonMonIndexMin) 
		tau = tau / 2.;
	else if (tau < tauAdjust)
		tau = min(tauHCinf, 1.01*tau);
	else
		tau = tauAdjust;

	tau = 1.e-8*(int)(1.e+8*tau);

	kg = 2 * h*h / (tau*Dg);
	bprg = 2. + kg;
}

//***********************************************************************************************************************************************************
//********************************* MAIN CYCLE ENDING *******************************************************************************************************
//***********************************************************************************************************************************************************

void Cycle_Ending()
{
	for (int i = 0; i <= NH; i++)
	{
		g0[i] = g1[i];
		n0[i] = n1[i];
		h0[i] = h1[i];
	}
}

//***********************************************************************************************************************************************************
//********************************* PROGRAM ENDING **********************************************************************************************************
//***********************************************************************************************************************************************************

void End()
{
	if (clock() / CLOCKS_PER_SEC >= 3600)
		printf("\n%s%ld%s%ld%s%ld%s", "The end. Total time: ", (clock() / CLOCKS_PER_SEC) / 3600, " h ", (clock() / CLOCKS_PER_SEC) / 60 - 60 * ((clock() / CLOCKS_PER_SEC) / 3600), " min ", (clock() / CLOCKS_PER_SEC) - 3600 * ((clock() / CLOCKS_PER_SEC) / 3600) - 60 * ((clock() / CLOCKS_PER_SEC) / 60 - 60 * ((clock() / CLOCKS_PER_SEC) / 3600)), " s.");
	else if (clock() / CLOCKS_PER_SEC >= 60)
		printf("\n%s%ld%s%ld%s", "The end. Total time: ", (clock() / CLOCKS_PER_SEC) / 60 - 60 * ((clock() / CLOCKS_PER_SEC) / 3600), " min ", (clock() / CLOCKS_PER_SEC) - 3600 * ((clock() / CLOCKS_PER_SEC) / 3600) - 60 * ((clock() / CLOCKS_PER_SEC) / 60 - 60 * ((clock() / CLOCKS_PER_SEC) / 3600)), " s.");
	else
		printf("\n%s%ld%s", "The end. Total time: ", (clock() / CLOCKS_PER_SEC) - 3600 * ((clock() / CLOCKS_PER_SEC) / 3600) - 60 * ((clock() / CLOCKS_PER_SEC) / 60 - 60 * ((clock() / CLOCKS_PER_SEC) / 3600)), " s.");
	//_getch();
	exit(0);
}

//********************************************************
//********* THE END
//********************************************************
